<?xml version="1.0" encoding="UTF-8"?>
<!--
| Copyright (c) 2017-2023 XMLmind Software. All rights reserved.
|
| Author: Hussein Shafie (hussein@xmlmind.com)
|
| This file is part of the XMLmind XML Editor project.
| For conditions of distribution and use, see the accompanying legal.txt file.
-->
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:fo="http://www.w3.org/1999/XSL/Format"
                xmlns:html="http://www.w3.org/1999/xhtml"
                xmlns:h="http://www.xmlmind.com/namespace/xsl/fo/helpers"
                exclude-result-prefixes="xs h"
                version="2.0">

  <!-- Common attributes ================================================= -->

  <xsl:template name="commonAttributes">
    <xsl:call-template name="idAttribute"/>
    <xsl:call-template name="localizationAttributes"/>
    <xsl:call-template name="xmlSpaceAttribute"/>
  </xsl:template>

  <xsl:template name="idAttribute">
    <!-- XSL-FO does not mandate its id to be an xs:ID. 
         It just says: "a unique identifier which does contain whitespace". -->
    <xsl:copy-of select="@id"/>
  </xsl:template>

  <xsl:template name="localizationAttributes">
    <!-- * HTML supports lang and xml:lang but gives priority to xml:lang.
         * XSL treats xml:lang as a shorthand and uses it to set the country and
         language properties. -->
    <xsl:choose>
      <xsl:when test="exists(@xml:lang)">
        <xsl:copy-of select="@xml:lang"/>
      </xsl:when>
      <xsl:when test="exists(@lang)">
        <xsl:attribute name="xml:lang" select="string(@lang)"/>
      </xsl:when>
      <xsl:otherwise>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="xmlSpaceAttribute">
    <xsl:if test="@xml:space eq 'preserve'">
      <xsl:attribute name="white-space">pre</xsl:attribute>
    </xsl:if>
  </xsl:template>

  <!-- Helpers =========================================================== -->

  <!-- resolveHref ========== -->

  <xsl:template name="resolveHref">
    <xsl:param name="href" select="''"/>

    <xsl:choose>
      <xsl:when test="matches($href, '^[a-zA-Z][a-zA-Z0-9.+-]*:/')">
        <xsl:value-of select="$href"/>
      </xsl:when>

      <xsl:otherwise>
        <xsl:variable name="baseList1" 
          select="ancestor-or-self::*/@xml:base[. ne '']/string()"/>
        <xsl:variable name="baseList2" select="($baseList1, $href)"/>

        <xsl:variable name="base" 
          select="if (exists(/html:html/html:head/html:base/@href))
                  then string((/html:html/html:head/html:base/@href)[1])
                  else ''"/>

        <xsl:variable name="docBaseURI" 
                      select="if ($base ne '')
                              then resolve-uri($base, base-uri(/))
                              else base-uri(/)"/>

        <xsl:variable name="baseList3" select="($docBaseURI, $baseList2)"/>

        <xsl:variable name="resolved" 
          select="h:doResolveHref($baseList3[1], subsequence($baseList3, 2))"/>

        <xsl:value-of select="$resolved"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:function name="h:doResolveHref" as="xs:string">
    <xsl:param name="base" as="xs:string"/>
    <xsl:param name="uris" as="xs:string*"/>

    <xsl:choose>
      <xsl:when test="empty($uris)">
        <xsl:sequence select="$base"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:variable name="newBase" select="resolve-uri($uris[1], $base)"/>
        <xsl:sequence 
          select="h:doResolveHref($newBase, subsequence($uris, 2))"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:function>

  <!-- asLengthOrPercent, asLength ========== -->

  <xsl:function name="h:asLengthOrPercent" as="xs:string">
    <xsl:param name="value" as="xs:string"/>

    <xsl:variable name="value2" select="normalize-space($value)"/>

    <xsl:choose>
      <xsl:when test="ends-with($value2, '%')">
        <xsl:sequence select="$value2"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:sequence select="h:asLength($value2)"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:function>

  <xsl:function name="h:asLength" as="xs:string">
    <xsl:param name="value" as="xs:string"/>

    <xsl:variable name="value2" select="normalize-space($value)"/>

    <xsl:variable name="px" as="xs:string?">
      <xsl:choose>
        <xsl:when test="ends-with($value2, 'px')">
          <xsl:variable name="value3"
                        select="substring-before($value2, 'px')"/>
                        
          <xsl:if test="number($value3) ge 0">
            <xsl:sequence select="$value3"/>
          </xsl:if>
        </xsl:when>
        <xsl:when test="number($value2) ge 0">
          <xsl:sequence select="$value2"/>
        </xsl:when>
      </xsl:choose>
    </xsl:variable>

    <xsl:choose>
      <xsl:when test="exists($px)">
        <xsl:variable name="pt" 
          select="(number($px) * 72.0) div number($screen-resolution)"/>

        <xsl:sequence select="concat($pt, 'pt')"/>
      </xsl:when>
      <xsl:otherwise>
        <!-- Examples: 12pt, auto, inherit, thick, center, smaller. -->
        <xsl:sequence select="$value2"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:function>

  <!-- isNumber ========== -->

  <xsl:function name="h:isNumber" as="xs:boolean">
    <xsl:param name="value" as="xs:string"/>

    <xsl:sequence 
      select="($value ne '' and string(number($value)) ne 'NaN')"/>
  </xsl:function>

  <!-- indexOfNode ========== -->

  <xsl:function name="h:indexOfNode" as="xs:integer?">
    <xsl:param name="nodeSequence" as="node()*"/>
    <xsl:param name="searchedNode" as="node()"/>

    <xsl:for-each select="$nodeSequence">
      <xsl:if test=". is $searchedNode">
        <xsl:sequence select="position()"/>
      </xsl:if>
    </xsl:for-each>
  </xsl:function>

  <!-- containsBlock ========== -->

  <xsl:function name="h:containsBlock" as="xs:boolean">
    <xsl:param name="container" as="element()"/>

    <xsl:sequence select="exists($container/html:h1 |
                                 $container/html:h2 |
                                 $container/html:h3 |
                                 $container/html:h4 |
                                 $container/html:h5 |
                                 $container/html:h6 |
                                 $container/html:hgroup |
                                 $container/html:div |
                                 $container/html:center |
                                 $container/html:section |
                                 $container/html:nav |
                                 $container/html:article |
                                 $container/html:aside |
                                 $container/html:main |
                                 $container/html:header |
                                 $container/html:footer |
                                 $container/html:blockquote |
                                 $container/html:address |
                                 $container/html:hr |
                                 $container/html:p |
                                 $container/html:pre |
                                 $container/html:ul |
                                 $container/html:ol |
                                 $container/html:dl |
                                 $container/html:dir |
                                 $container/html:menu |
                                 $container/html:table |
                                 $container/html:figure |
                                 $container/html:form |
                                 $container/html:fieldset)"/>
  </xsl:function>

</xsl:stylesheet>
