<?xml version="1.0" encoding="UTF-8"?>
<!--
| Copyright (c) 2017 XMLmind Software. All rights reserved.
|
| Author: Hussein Shafie
|
| This file is part of the XMLmind Ebook Compiler project.
| For conditions of distribution and use, see the accompanying LEGAL.txt file.
-->
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:u="http://www.xmlmind.com/namespace/ebookc"
                exclude-result-prefixes="u xs"
                version="2.0">

  <!-- has =============================================================== -->

  <xsl:function name="u:has" as="xs:boolean">
    <xsl:param name="attr" as="attribute()?"/> <!--T ypically @class -->
    <xsl:param name="list" as="xs:string+"/>

    <xsl:sequence select="if (empty($attr))
                          then false()
                          else (tokenize($attr, '\s+') = $list)"/>
  </xsl:function>

  <!-- matches =========================================================== -->

  <xsl:function name="u:matches" as="xs:boolean">
    <xsl:param name="attr" as="attribute()?"/> <!-- Typically @class -->
    <xsl:param name="prefix" as="xs:string"/> <!-- '' is OK -->
    <xsl:param name="suffix" as="xs:string"/> <!-- '' is OK -->

    <xsl:sequence select="if (empty($attr))
                          then false()
                          else exists(for $t in tokenize($attr, '\s+')
                                      return if (starts-with($t, $prefix) and 
                                                 ends-with($t, $suffix))
                                             then true()
                                             else ())"/>
  </xsl:function>

  <!-- basename ========================================================== -->

  <xsl:function name="u:basename" as="xs:string">
    <xsl:param name="path" as="xs:string"/>

    <xsl:sequence select="if ($path ne '') 
                          then tokenize($path, '/')[last()]
                          else ''" />
  </xsl:function>

  <!-- extension ========================================================= -->

  <xsl:function name="u:extension" as="xs:string">
    <xsl:param name="path" as="xs:string"/>

    <xsl:variable name="basename" select="u:basename($path)" />
    <xsl:sequence select="if ($basename ne '') 
                          then tokenize($basename, '\.')[last()]
                          else ''" />
  </xsl:function>

  <!-- indexOfNode ======================================================= -->

  <xsl:function name="u:indexOfNode" as="xs:integer*">
    <xsl:param name="nodeSequence" as="node()*"/>
    <xsl:param name="searchedNode" as="node()"/>

    <xsl:for-each select="$nodeSequence">
      <xsl:if test=". is $searchedNode">
        <xsl:sequence select="position()"/>
      </xsl:if>
    </xsl:for-each>
  </xsl:function>

  <!-- localize ========================================================== -->

  <xsl:param name="messageLangs"
    select="'cs', 'de', 'en', 'es', 'fr', 'it', 'pl', 'ru', 'nn', 'nb', 'ja'"/>

  <xsl:template name="localize">
    <xsl:param name="message" select="''"/>

    <xsl:variable name="lang">
      <xsl:call-template name="lang"/>
    </xsl:variable>

    <xsl:variable name="lang2" select="lower-case($lang)"/>

    <xsl:variable name="messageLang">
      <xsl:choose>
        <xsl:when test="index-of($messageLangs, $lang2) ge 1">
          <xsl:value-of select="$lang2"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:variable name="lang3" select="substring-before($lang2, '-')"/>
          <xsl:choose>
            <xsl:when test="index-of($messageLangs, $lang3) ge 1">
              <xsl:value-of select="$lang3"/>
            </xsl:when>
            <xsl:otherwise>en</xsl:otherwise>
          </xsl:choose>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>

    <xsl:variable name="messageFile"
      select="concat('messages/messages_', $messageLang, '.xml')"/>

    <!-- Documents loaded using doc() are cached. -->
    <xsl:variable name="properties"
                  select="if (doc-available($messageFile)) 
                          then doc($messageFile)/properties
                          else ()"/>

    <xsl:choose>
      <xsl:when test="exists($properties/entry[@key eq $message])">
        <xsl:value-of select="string($properties/entry[@key eq $message])"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:variable name="enProperties"
                      select="doc('messages/messages.xml')/properties"/>
        <xsl:choose>
          <xsl:when test="exists($enProperties/entry[@key eq $message])">
            <xsl:value-of
              select="string($enProperties/entry[@key eq $message])"/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:value-of select="$message"/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="lang">
    <xsl:variable name="lang"
      select="lower-case(if (ancestor-or-self::*/@lang) 
                         then string((ancestor-or-self::*/@lang)[last()])
                         else 'en')"/>
    
    <!-- Fix a common error. -->
    <xsl:variable name="lang2" select="translate($lang, '_', '-')"/>

    <!-- Get rid of variant. Example: fr-CA-Windows. -->
    <xsl:choose>
      <xsl:when test="string-length($lang2) gt 5">
        <xsl:value-of select="substring($lang2, 1, 5)"/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$lang2"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- Function variant -->

  <xsl:function name="u:localize" as="xs:string">
    <xsl:param name="message" as="xs:string"/>
    <xsl:param name="context" as="element()"/>

    <xsl:for-each select="$context">
      <xsl:call-template name="localize">
        <xsl:with-param name="message" select="$message"/>
      </xsl:call-template>
    </xsl:for-each>
  </xsl:function>

</xsl:stylesheet>
