<?xml version="1.0" encoding="UTF-8"?>
<!--
| Copyright (c) 2017 XMLmind Software. All rights reserved.
|
| Author: Hussein Shafie
|
| This file is part of the XMLmind Ebook Compiler project.
| For conditions of distribution and use, see the accompanying LEGAL.txt file.
-->
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:opf="http://www.idpf.org/2007/opf"
                xmlns:dc="http://purl.org/dc/elements/1.1/"
                xmlns:mml="http://www.w3.org/1998/Math/MathML"
                xmlns:svg="http://www.w3.org/2000/svg"
                xmlns:u="http://www.xmlmind.com/namespace/ebookc"
                xpath-default-namespace="http://www.w3.org/1999/xhtml"
                exclude-result-prefixes="xs mml svg u"
                version="2.0">

  <xsl:output name="contentOutputFormat"
              method="xml" encoding="UTF-8" indent="yes"/>

  <xsl:template name="generateContent">
    <xsl:result-document href="{resolve-uri('content.opf', base-uri())}"
                         format="contentOutputFormat">
      <opf:package version="3.0" unique-identifier="__ID">
        <xsl:call-template name="generateMetadata"/>

        <xsl:variable name="items" 
          select="//li[u:has(@class, 'role-booklist-entry') and
                       ./a[u:matches(@class, '', '-ref') and 
                           exists(@href) and not(contains(@href, '#'))]]"/>

        <xsl:call-template name="generateManifest">
          <xsl:with-param name="items" select="$items"/>
        </xsl:call-template>

        <xsl:call-template name="generateSpine">
          <xsl:with-param name="items" select="$items"/>
        </xsl:call-template>
      </opf:package>
    </xsl:result-document>
  </xsl:template>

  <!-- ========== metadata ========== -->

  <xsl:template name="generateMetadata">
    <opf:metadata>
      <dc:identifier id="__ID">
        <xsl:value-of select="$epubIdentifier"/>
      </dc:identifier>

      <dc:title>
        <xsl:value-of select="normalize-space(./head/title)"/>
      </dc:title>

      <dc:language>
        <xsl:value-of select="if (exists(@lang)) then @lang else 'en'"/>
      </dc:language>

      <!-- Translate author if any to dc:creator. -->
      <xsl:if test="empty(./head/meta[@name eq 'dc.creator'])">
        <xsl:for-each select="./head/meta[@name eq 'author']">
          <xsl:variable name="content" select="normalize-space(@content)"/>

          <xsl:if test="$content ne ''">
            <dc:creator>
              <xsl:value-of select="$content"/>
            </dc:creator>
          </xsl:if>
        </xsl:for-each>
      </xsl:if>

      <!-- Translate description if any to dc:description. -->
      <xsl:if test="empty(./head/meta[@name eq 'dc.description'])">
        <xsl:for-each select="./head/meta[@name eq 'description']">
          <xsl:variable name="content" select="normalize-space(@content)"/>

          <xsl:if test="$content ne ''">
            <dc:description>
              <xsl:value-of select="$content"/>
            </dc:description>
          </xsl:if>
        </xsl:for-each>
      </xsl:if>

      <xsl:for-each select="./head/meta[starts-with(@name, 'dc.')]">
        <xsl:variable name="name" select="substring-after(@name, 'dc.')"/>
        <xsl:variable name="content" select="normalize-space(@content)"/>

        <xsl:if test="$name = ('contributor', 'coverage', 'creator', 'date', 
                               'description', 'format', 'publisher', 'relation',
                               'rights', 'source', 'subject', 'type') and
                      $content ne ''">
          <xsl:element name="{concat('dc:', $name)}"
                       namespace="http://purl.org/dc/elements/1.1/">
            <xsl:value-of select="$content"/>
          </xsl:element>
        </xsl:if>
      </xsl:for-each>

      <xsl:for-each select="./head/meta[starts-with(@name, 'dcterms.')]">
        <xsl:variable name="name" select="substring-after(@name, 'dcterms.')"/>
        <xsl:variable name="content" select="normalize-space(@content)"/>

        <xsl:if test="$content ne ''">
          <!-- "dcterms:" is a reserved prefix. -->
          <opf:meta property="{concat('dcterms:', $name)}">
            <xsl:value-of select="$content"/>
          </opf:meta>
        </xsl:if>
      </xsl:for-each>

      <!-- At least one opf:meta is required. -->
      <xsl:if test="empty(./head/meta[@name eq 'dcterms.modified'])">
        <opf:meta property="dcterms:modified">
          <xsl:value-of
            select="format-dateTime(
                      adjust-dateTime-to-timezone(current-dateTime(), 
                                                  xs:dayTimeDuration('PT0S')),
                     '[Y0001]-[M01]-[D01]T[H01]:[m01]:[s01]Z', 'en',(),'US')"/>
        </opf:meta>
      </xsl:if>

      <xsl:if test="$coverImageName ne '' and $epub2-compatible eq 'yes'">
        <opf:meta name="cover" content="cover-image"/>
      </xsl:if>
    </opf:metadata>
  </xsl:template>

  <!-- ========== manifest ========== -->

  <xsl:template name="generateManifest">
    <xsl:param name="items" select="()"/>

    <opf:manifest>
      <!-- nav -->

      <opf:item id="__TOC" properties="nav" 
                href="__toc__.xhtml" media-type="application/xhtml+xml"/>

      <xsl:if test="$epub2-compatible eq 'yes'">
        <opf:item id="__TOC_NCX" href="toc.ncx" 
                  media-type="application/x-dtbncx+xml" />
      </xsl:if>

      <!-- cover-image -->

      <xsl:if test="$coverImageName ne ''">
        <xsl:if test="$epub2-compatible eq 'yes'">
          <opf:item id="cover" href="__cover__.xhtml" 
                    media-type="application/xhtml+xml"/>
        </xsl:if>
        <!-- Otherwise, with EPUB3, an XHTML page containing the cover image
             is not useful. -->

        <opf:item id="cover-image" properties="cover-image" 
                  href="{$coverImageName}" 
                  media-type="{u:mediaType($coverImageName)}"/>
      </xsl:if>

      <!-- XHTML files -->

      <xsl:for-each select="$items">
        <xsl:variable name="href"
          select="string(./a[u:matches(@class, '', '-ref')]/@href)"/>

        <opf:item id="{concat('item_', @id)}" href="{$href}"
                  media-type="application/xhtml+xml">
          <xsl:call-template name="itemProperties">
            <xsl:with-param name="href" select="$href"/>
          </xsl:call-template>
        </opf:item>
      </xsl:for-each>

      <!-- Resource (e.g. images) files -->

      <xsl:for-each select="/html/head/link[@rel eq 'resource' and
                                            (empty(@type) or 
                                             @type ne 'inode/directory')]">
        <xsl:variable name="href" select="string(@href)"/>
        <xsl:variable name="type"
          select="if (empty(@type))
                  then 'application/octet-stream'
                  else string(@type)"/>

        <opf:item id="{concat('res_', position())}" href="{$href}"
                  media-type="{$type}">
          <xsl:if test="$type eq 'application/xhtml+xml'">
            <xsl:call-template name="itemProperties">
              <xsl:with-param name="href" select="$href"/>
            </xsl:call-template>
          </xsl:if>
        </opf:item>
      </xsl:for-each>
    </opf:manifest>
  </xsl:template>

  <xsl:template name="itemProperties">
    <xsl:param name="href" select="''"/>

    <xsl:variable name="page"
                  select="doc(resolve-uri($href, base-uri()))"/>

    <xsl:variable name="properties">
      <!-- SVG embedded by inclusion or by reference. -->
      <xsl:if 
        test="exists($page//svg:svg) or
              exists($page//*[((self::img or self::iframe) and 
                              (ends-with(@src, '.svg') or 
                               ends-with(@src, '.svgz'))) or 
                               (self::object and 
                                (ends-with(@data, '.svg') or 
                                 ends-with(@data, '.svgz')))])">
        <xsl:text> svg</xsl:text>
      </xsl:if>

      <xsl:if test="exists($page//mml:math)">
        <xsl:text> mathml</xsl:text>
      </xsl:if>

      <xsl:if test="exists($page//*[self::script or 
                                    self::form or 
                                    self::label or 
                                    self::input or 
                                    self::button or 
                                    self::select or 
                                    self::textarea or 
                                    self::keygen or 
                                    self::output or 
                                    self::progress or 
                                    self::meter or 
                                    self::fieldset or 
                                    self::legend])">
        <xsl:text> scripted</xsl:text>
      </xsl:if>
    </xsl:variable>

    <xsl:variable name="properties2" select="normalize-space($properties)"/>
    <xsl:if test="$properties2 ne ''">
      <xsl:attribute name="properties" select="$properties2"/>
    </xsl:if>
  </xsl:template>

  <xsl:function name="u:mediaType">
    <xsl:param name="path"/>

    <xsl:variable name="path2" select="lower-case($path)"/>
    <xsl:sequence
      select="if (ends-with($path2, '.gif')) then 'image/gif'
              else if (ends-with($path2, '.jpg') or
                       ends-with($path2, '.jpeg')) then 'image/jpeg'
              else if (ends-with($path2, '.png')) then 'image/png'
              else if (ends-with($path2, '.svg') or 
                       ends-with($path2, '.svgz')) then 'image/svg+xml'

              else 'application/octet-stream'"/>
  </xsl:function>

  <!-- ========== spine ========== -->

  <xsl:template name="generateSpine">
    <xsl:param name="items" select="()"/>

    <opf:spine>
      <xsl:if test="$epub2-compatible eq 'yes'">
        <xsl:attribute name="toc" select="'__TOC_NCX'"/>
      </xsl:if>
      <!-- But no need to reference __toc__.xhtml here. -->

      <xsl:if test="$coverImageName ne '' and $epub2-compatible eq 'yes'">
        <opf:itemref idref="cover" linear="no"/>
      </xsl:if>

      <xsl:for-each select="$items">
        <opf:itemref idref="{concat('item_', @id)}" />
      </xsl:for-each>
    </opf:spine>
  </xsl:template>

</xsl:stylesheet>
