<?xml version="1.0" encoding="UTF-8"?>
<!--
| Copyright (c) 2017-2024 XMLmind Software. All rights reserved.
|
| Author: Hussein Shafie (hussein@xmlmind.com)
|
| This file is part of the XMLmind Ebook Compiler project.
| For conditions of distribution and use, see the accompanying LEGAL.txt file.
-->
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:fo="http://www.w3.org/1999/XSL/Format"
                xmlns:html="http://www.w3.org/1999/xhtml"
                xmlns:h="http://www.xmlmind.com/namespace/xsl/fo/helpers"
                xmlns:u="http://www.xmlmind.com/namespace/ebookc"
                xmlns:Date="java:com.xmlmind.ebookc.xslt.Date"
                exclude-result-prefixes="xs html h u Date"
                version="2.0">

  <xsl:import href="../common/commonUtil.xsl"/>
  <xsl:import href="params.xsl"/>
  <xsl:import href="helpers.xsl"/>
  <xsl:import href="cssStyles.xsl"/>
  <xsl:import href="pagination.xsl"/>
  <xsl:import href="body.xsl"/>
  <xsl:import href="table.xsl"/>
  <xsl:import href="pdfOutline.xsl"/>

  <!-- Set foProcessor =================================================== -->

  <!-- ``System parameters'' automatically specified by the application
       hosting ebookc. -->
  <xsl:param name="outputFormat" select="''"/>

  <!-- If the name of the system property is in the null namespace, Saxon
       returns the value of the Java system property whose name matches the
       local name. -->

  <xsl:param name="foProcessor"
    select="if ($outputFormat = ('pdf', 'ps')) 
            then
                if (contains(system-property('XSL_FO_PROCESSORS'), 'AHF')) 
                then 'AHF'
                else
                  if (contains(system-property('XSL_FO_PROCESSORS'), 'XEP')) 
                  then 'XEP'
                  else
                      if (contains(system-property('XSL_FO_PROCESSORS'), 'FOP'))
                      then 'FOP' else ''
            else
                if ($outputFormat = ('rtf', 'wml', 'docx', 'odt'))
                then
                    if (contains(system-property('XSL_FO_PROCESSORS'), 'XFC')) 
                    then 'XFC' else ''
                else ''"/>

  <!-- Output ============================================================ -->

  <xsl:output method="xml" encoding="UTF-8" indent="no"/>

  <xsl:preserve-space elements="html:pre html:style html:script"/>

  <!-- Override some of the default styles =============================== -->

  <xsl:attribute-set name="h6">
    <xsl:attribute name="font-size">1em</xsl:attribute>
    <xsl:attribute name="space-before">1em</xsl:attribute>
    <xsl:attribute name="space-after">1em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:p">
    <fo:block xsl:use-attribute-sets="p">
      <!--LEGACY-->
      <xsl:call-template name="alignAttribute"/>
      <xsl:call-template name="cssStyles"/>

      <xsl:if test="$foProcessor eq 'XEP' and
                    (parent::html:li or 
                     parent::html:div[u:has(@class, 'role-footnote')]) and
                    empty(following-sibling::*)">
        <!-- Minimal workaround for implementation limit in XEP:
             space-after.conditionality="discard" is not implemented,
             fallback value is "retain".
             space-after each paragraph is stacked instead of being 
             overlapped. As a result, there is much too space between 
             list items. 
             There is no such problem with FOP or XFC. -->
        <xsl:attribute name="space-after">0em</xsl:attribute>
      </xsl:if>

      <xsl:call-template name="commonAttributes"/>
      <xsl:apply-templates/>
    </fo:block>
  </xsl:template>

  <!-- The root element ================================================== -->

  <xsl:template match="html:html">
    <fo:root>
      <xsl:call-template name="localizationAttributes"/>

      <xsl:call-template name="layoutMasterSet"/>

      <xsl:if test="$pdf-outline eq 'yes'">
        <xsl:for-each select="//html:section[u:has(@class, 'role-toc')]">
          <xsl:choose>
            <xsl:when test="$foProcessor eq 'XEP'">
              <xsl:call-template name="xepPDFOutline"/>
            </xsl:when>
            <xsl:when test="$foProcessor eq 'FOP' or $foProcessor eq 'AHF'">
              <xsl:call-template name="xsl11PDFOutline"/>
            </xsl:when>
          </xsl:choose>
        </xsl:for-each>
      </xsl:if>

      <xsl:variable name="documentDivisions" 
        select="//html:section[u:has(@class, 
                                     ('role-book', 'role-toc', 'role-lot', 
                                      'role-lof', 'role-lox', 'role-loe', 
                                      'role-index', 'role-part', 
                                      'role-appendices', 'role-appendix', 
                                      'role-chapter')) or
                               (u:has(@class, 'role-section1') and 
                                parent::html:div[@class =
                                                 ('role-frontmatter', 
                                                  'role-backmatter')])]"/>

      <xsl:variable name="documentTitle"
                    select="normalize-space(/html:html/html:head/html:title)"/>

      <xsl:variable name="documentDate">
        <xsl:call-template name="documentDate"/>
      </xsl:variable>

      <xsl:apply-templates
        select="./html:body/html:section[u:has(@class, 'role-book')]">
        <xsl:with-param name="documentDivisions" 
                        select="$documentDivisions" tunnel="yes"/>
        <xsl:with-param name="documentTitle" 
                        select="$documentTitle" tunnel="yes"/>
        <xsl:with-param name="documentDate" 
                        select="$documentDate" tunnel="yes"/>
      </xsl:apply-templates>
    </fo:root>
  </xsl:template>

  <xsl:template name="documentDate">
    <xsl:param name="defaulted" select="true()"/>

    <xsl:variable name="date1"
       select="/html:html/html:head/html:meta[@name eq 'dc.date']"/>
    <xsl:variable name="date2"
      select="/html:html/html:head/html:meta[@name eq 'dcterms.issued']"/>
    <xsl:variable name="date3"
      select="/html:html/html:head/html:meta[@name eq 'dcterms.modified']"/>
    <xsl:variable name="date4"
      select="/html:html/html:head/html:meta[@name eq 'dcterms.created']"/>

    <xsl:variable name="date">
      <xsl:choose>
        <xsl:when test="normalize-space($date1[1]/@content) ne ''">
          <xsl:value-of select="normalize-space($date1[1]/@content)"/>
        </xsl:when>
        <xsl:when test="normalize-space($date2[1]/@content) ne ''">
          <xsl:value-of select="normalize-space($date2[1]/@content)"/>
        </xsl:when>
        <xsl:when test="normalize-space($date3[1]/@content) ne ''">
          <xsl:value-of select="normalize-space($date3[1]/@content)"/>
        </xsl:when>
        <xsl:when test="normalize-space($date4[1]/@content) ne ''">
          <xsl:value-of select="normalize-space($date4[1]/@content)"/>
        </xsl:when>
      </xsl:choose>
    </xsl:variable>

    <xsl:variable name="lang">
      <xsl:call-template name="lang"/>
    </xsl:variable>

    <xsl:value-of select="if ($date ne '')
                          then Date:format($date, $lang)
                          else if ($defaulted)
                          then Date:format('', $lang)
                          else ''"/>
  </xsl:template>

  <!-- role-book ========================================================= -->

  <xsl:attribute-set name="base-style">
    <xsl:attribute name="font-family" select="$font-family"/>
    <xsl:attribute name="font-size" select="$base-font-size"/>
    <xsl:attribute name="line-height" select="$base-line-height"/>
    <xsl:attribute name="text-align"
                  select="if ($justified eq 'yes') then 'justify' else 'left'"/>
  </xsl:attribute-set>

  <!-- IMPORTANT NOTE:
       The generated XSL-FO is expected to be styled using cssStyles.
       That's why the role-XXX attribute-sets are are unimportant 
       hence not defined for all elements. -->

  <xsl:attribute-set name="role-book">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-book')]">
    <xsl:variable name="subsections" 
                  select="html:div[u:has(@class, 'role-frontmatter')] |
                          html:div[u:has(@class, 'role-backmatter')] |
                          html:section[u:has(@class, 'role-part')] |
                          html:section[u:has(@class, 'role-chapter')] |
                          html:section[u:has(@class, 'role-appendices')] |
                          html:section[u:has(@class, 'role-appendix')]"/>

    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-book">
          <xsl:call-template name="rootBlockCSSStyles"/>
          <xsl:call-template name="commonAttributes"/>

          <xsl:apply-templates select="* except $subsections"/>
        </fo:block>
      </fo:flow>
    </fo:page-sequence>

    <xsl:apply-templates select="$subsections"/>
  </xsl:template>

  <xsl:template name="rootBlockCSSStyles">
    <!-- body is never traversed but generally has useful CSS styles. -->
    <xsl:for-each select="ancestor::html:body">
      <xsl:call-template name="cssStyles"/>
    </xsl:for-each>

    <xsl:call-template name="cssStyles"/>
  </xsl:template>

  <!-- role-book-title-div processed normally like all 
       the other elements which are not processed in this stylesheet. ==== -->

  <!-- role-book-title =================================================== -->

  <xsl:attribute-set name="title-style">
    <xsl:attribute name="keep-with-next.within-column">always</xsl:attribute>
  </xsl:attribute-set>

  <xsl:attribute-set name="role-book-title"
                     use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-book-title')]">
    <fo:block xsl:use-attribute-sets="role-book-title">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <xsl:template name="processBlock">
    <xsl:call-template name="cssStyles"/>
    <xsl:call-template name="commonAttributes"/>
    <xsl:apply-templates/>
  </xsl:template>

  <xsl:template name="processInline">
    <xsl:call-template name="cssStyles">
      <xsl:with-param name="select" select="'inline'"/>
    </xsl:call-template>
    <xsl:call-template name="commonAttributes"/>
    <xsl:apply-templates/>
  </xsl:template>

  <!-- Transparent role-frontmatter,-backmatter ========================== -->

  <xsl:template match="html:div[@class = ('role-frontmatter', 
                                          'role-backmatter')]">
    <xsl:apply-templates/>
  </xsl:template>

  <!-- role-toc ========================================================== -->

  <xsl:attribute-set name="role-toc">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-toc')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-toc">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <xsl:template name="processRootBlock">
    <xsl:call-template name="rootBlockCSSStyles"/>
    <xsl:call-template name="commonAttributes"/>
    <xsl:apply-templates/>
  </xsl:template>

  <xsl:template name="endOfFrontBackMatter">
    <xsl:if test="empty(following-sibling::*)">
      <xsl:choose>
        <xsl:when test="exists(parent::html:div[u:has(@class, 'role-frontmatter')])">
          <fo:block id="__EOFM" hyphenate="false"/>
        </xsl:when>
        <xsl:when test="exists(parent::html:div[u:has(@class, 'role-backmatter')])">
          <fo:block id="__EOBM" hyphenate="false"/>
        </xsl:when>
      </xsl:choose>
    </xsl:if>
  </xsl:template>

  <!-- role-toc-title ==================================================== -->

  <xsl:attribute-set name="role-toc-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-toc-title')]">
    <fo:block xsl:use-attribute-sets="role-toc-title">
      <xsl:call-template name="processDivisionTitle"/>
    </fo:block>
  </xsl:template>

  <xsl:template name="processDivisionTitle">
    <xsl:call-template name="xfcOutlineLevel"/>
    <xsl:call-template name="cssStyles"/>
    <xsl:call-template name="commonAttributes"/>

    <fo:marker marker-class-name="division-title">
      <xsl:value-of select="normalize-space(.)"/>
    </fo:marker>

    <xsl:apply-templates/>
  </xsl:template>

  <xsl:template name="xfcOutlineLevel">
    <xsl:param name="level" select="1"/>
    
    <xsl:if test="$foProcessor eq 'XFC'">
      <xsl:variable name="level2" 
        select="if (u:has(@class,
                          ('role-section1-title', 'role-section2-title',
                           'role-section3-title', 'role-section4-title',
                           'role-section5-title', 'role-section6-title',
                           'role-section7-title', 'role-section8-title',
                           'role-section9-title')) and
                    exists(ancestor::*[u:has(@class, 
                                            ('role-chapter','role-appendix'))]))
              then $level + 1
              else $level"/>

      <xsl:variable name="level3" 
        select="if (u:has(@class, 
                          ('role-chapter-title', 'role-appendix-title',
                           'role-section1-title', 'role-section2-title',
                           'role-section3-title', 'role-section4-title',
                           'role-section5-title', 'role-section6-title',
                           'role-section7-title', 'role-section8-title',
                           'role-section9-title')) and
                    exists(ancestor::*[u:has(@class, 
                                             ('role-part','role-appendices'))]))
              then $level2 + 1
              else $level2"/>

      <xsl:if test="$level3 le 9">
        <xsl:attribute name="xfc:outline-level" 
          xmlns:xfc="http://www.xmlmind.com/foconverter/xsl/extensions"
          select="$level3"/>
      </xsl:if>
    </xsl:if>
  </xsl:template>

  <!-- role-booklist-entries ============================================= -->
  <!-- Translated to nested blocks and not to a list-block. -->

  <xsl:attribute-set name="role-booklist-entries"
                     use-attribute-sets="block-style">
  </xsl:attribute-set>

  <xsl:template match="html:ul[u:has(@class, 'role-booklist-entries')]">
    <fo:block xsl:use-attribute-sets="role-booklist-entries">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <xsl:attribute-set name="nested-booklist-entries"
    use-attribute-sets="role-booklist-entries compact-block-style">
    <xsl:attribute name="margin-left">1.5em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:ul[u:has(@class, 'role-booklist-entries')]//
                       html:ul[u:has(@class, 'role-booklist-entries')]"
                priority="2">
    <fo:block xsl:use-attribute-sets="nested-booklist-entries">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-booklist-entry =============================================== -->

  <xsl:attribute-set name="role-booklist-entry"
                     use-attribute-sets="compact-block-style">
    <xsl:attribute name="text-align-last">justify</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:li[u:has(@class, 'role-booklist-entry')]">
    <fo:block xsl:use-attribute-sets="role-booklist-entry">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- a-*-ref inside a booklist ========================================= -->

  <xsl:attribute-set name="booklist-entry-link-style">
  </xsl:attribute-set>

  <xsl:template match="html:li[u:has(@class, 'role-booklist-entry')] /
                       html:a[u:matches(@class, 'role-', '-ref')]"
                priority="2">
    <xsl:variable name="ref" select="substring-after(@href, '#')"/>

    <fo:inline xsl:use-attribute-sets="booklist-entry-link-style">
      <xsl:call-template name="cssStyles">
        <xsl:with-param name="select" select="'inline'"/>
      </xsl:call-template>
      <xsl:call-template name="commonAttributes"/>

      <fo:basic-link internal-destination="{$ref}">
        <xsl:apply-templates/>
      </fo:basic-link>
      <xsl:text> </xsl:text>
      <fo:leader leader-pattern="dots"
                 keep-with-next.within-line="always">
        <xsl:if test="$foProcessor eq 'XFC'">
          <xsl:attribute
            xmlns:xfc="http://www.xmlmind.com/foconverter/xsl/extensions"
            name="xfc:tab-position">-30pt</xsl:attribute>
          <xsl:attribute
            xmlns:xfc="http://www.xmlmind.com/foconverter/xsl/extensions"
            name="xfc:tab-align">right</xsl:attribute>
        </xsl:if>
      </fo:leader>
      <xsl:text> </xsl:text>
      <fo:page-number-citation ref-id="{$ref}"/>
    </fo:inline>
  </xsl:template>

  <!-- role-lot ========================================================== -->

  <xsl:attribute-set name="role-lot">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-lot')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-lot">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-lot-title ==================================================== -->

  <xsl:attribute-set name="role-lot-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-lot-title')]">
    <fo:block xsl:use-attribute-sets="role-lot-title">
      <xsl:call-template name="processDivisionTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-lof ========================================================== -->

  <xsl:attribute-set name="role-lof">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-lof')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-lof">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-lof-title ==================================================== -->

  <xsl:attribute-set name="role-lof-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-lof-title')]">
    <fo:block xsl:use-attribute-sets="role-lof-title">
      <xsl:call-template name="processDivisionTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-lox ========================================================== -->

  <xsl:attribute-set name="role-lox">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-lox')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-lox">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-lox-title ==================================================== -->

  <xsl:attribute-set name="role-lox-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-lox-title')]">
    <fo:block xsl:use-attribute-sets="role-lox-title">
      <xsl:call-template name="processDivisionTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-loe ========================================================== -->

  <xsl:attribute-set name="role-loe">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-loe')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-loe">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-loe-title ==================================================== -->

  <xsl:attribute-set name="role-loe-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-loe-title')]">
    <fo:block xsl:use-attribute-sets="role-loe-title">
      <xsl:call-template name="processDivisionTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-index ======================================================== -->

  <xsl:attribute-set name="role-index">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-index')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <!-- Must be direct child of the fo:flow to apply span="all". -->
        <xsl:variable name="title"
                      select="./html:h1[u:has(@class, 'role-index-title')]"/>
        <xsl:for-each select="$title">
          <xsl:call-template name="processIndexTitle"/>
        </xsl:for-each>

        <fo:block xsl:use-attribute-sets="base-style role-index">
          <xsl:call-template name="rootBlockCSSStyles"/>
          <xsl:call-template name="commonAttributes"/>

          <xsl:apply-templates select="* except $title">
            <xsl:with-param name="pageNumberSepar" tunnel="yes">
              <xsl:call-template name="localize">
                <xsl:with-param name="message" select="'pageNumberSeparator'"/>
              </xsl:call-template>
            </xsl:with-param>
            <xsl:with-param name="pageRangeSepar" tunnel="yes">
              <xsl:call-template name="localize">
                <xsl:with-param name="message" select="'pageRangeSeparator'"/>
              </xsl:call-template>
            </xsl:with-param>
          </xsl:apply-templates>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-index-title ================================================== -->

  <xsl:attribute-set name="role-index-title" use-attribute-sets="h1">
  </xsl:attribute-set>

  <xsl:template name="processIndexTitle">
    <fo:block xsl:use-attribute-sets="role-index-title">
      <xsl:call-template name="xfcOutlineLevel"/>
      <xsl:call-template name="cssStyles"/>
      <xsl:call-template name="commonAttributes"/>

      <xsl:if test="$foProcessor eq 'XEP'">
        <xsl:attribute name="span">all</xsl:attribute>
      </xsl:if>
      <!-- span="all" does not work well, if at all, for the other XSL-FO
           processors (including AHF!). -->

      <fo:marker marker-class-name="division-title">
        <xsl:value-of select="normalize-space(.)"/>
      </fo:marker>

      <xsl:apply-templates/>
    </fo:block>
  </xsl:template>

  <!-- role-index-toc not useful in print form =========================== -->

  <xsl:template match="html:nav[u:has(@class, 'role-index-toc')]"/>

  <!-- role-index-div-title ============================================== -->

  <xsl:attribute-set name="role-index-div-title" use-attribute-sets="h2">
    <xsl:attribute name="space-after">0em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:h2[u:has(@class, 'role-index-div-title')]">
    <fo:block xsl:use-attribute-sets="role-index-div-title">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-index-toc-ref not useful in print form ======================= -->

  <xsl:template match="html:a[u:has(@class, 'role-index-toc-ref')]"/>

  <!-- role-index-entries ================================================= -->
  <!-- Translated to nested blocks and not to a list-block. -->

  <xsl:attribute-set name="role-index-entries"
                     use-attribute-sets="block-style">
  </xsl:attribute-set>

  <xsl:template match="html:ul[u:has(@class, 'role-index-entries')]">
    <fo:block xsl:use-attribute-sets="role-index-entries">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <xsl:attribute-set name="nested-index-entries"
    use-attribute-sets="role-index-entries compact-block-style">
    <xsl:attribute name="margin-left">2em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:ul[u:has(@class, 'role-index-entries')]//
                       html:ul[u:has(@class, 'role-index-entries')]"
                priority="2">
    <fo:block xsl:use-attribute-sets="nested-index-entries">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-index-anchor ================================================= -->

  <xsl:key name="indexAnchorRefs" 
           match="//html:a[u:has(@class, 'role-index-anchor-ref')]" 
           use="substring-after(@href, '#')"/>

  <xsl:template match="html:a[u:has(@class, 'role-index-anchor')]">
    <xsl:variable name="anchorRef" 
                  select="key('indexAnchorRefs', @id)"/>

    <xsl:choose>
      <xsl:when test="$foProcessor = ('XEP', 'AHF')">
        <xsl:for-each select="$anchorRef">
          <xsl:variable name="key"
            select="string((ancestor::html:li[u:has(@class, 'role-index-entry')])[last()]/@id)"/>

          <xsl:choose>
            <xsl:when test="parent::html:span[u:has(@class, 'role-index-range')]">
              <xsl:variable name="rangeId" 
                select="concat($key, '__',
                               replace(parent::*, '[^a-zA-Z0-9]', '_', 's'))"/>

              <xsl:choose>
                <xsl:when test="empty(preceding-sibling::*)">
                  <xsl:choose>
                    <xsl:when test="$foProcessor eq 'XEP'">
                      <rx:begin-index-range rx:key="{$key}" id="{$rangeId}"
                        xmlns:rx="http://www.renderx.com/XSL/Extensions"/>
                    </xsl:when>
                    <xsl:otherwise>
                      <fo:index-range-begin index-key="{$key}" id="{$rangeId}"/>
                    </xsl:otherwise>
                  </xsl:choose>
                </xsl:when>

                <xsl:otherwise>
                  <xsl:choose>
                    <xsl:when test="$foProcessor eq 'XEP'">
                      <rx:end-index-range ref-id="{$rangeId}"
                        xmlns:rx="http://www.renderx.com/XSL/Extensions"/>
                    </xsl:when>
                    <xsl:otherwise>
                      <fo:index-range-end ref-id="{$rangeId}"/>
                    </xsl:otherwise>
                  </xsl:choose>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:when>

            <xsl:otherwise>
              <xsl:choose>
                <xsl:when test="$foProcessor eq 'XEP'">
                  <fo:inline rx:key="{$key}"
                             xmlns:rx="http://www.renderx.com/XSL/Extensions"/>
                </xsl:when>
                <xsl:otherwise>
                  <fo:inline index-key="{$key}"/>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </xsl:when>

      <xsl:otherwise>
        <fo:inline hyphenate="false">
          <xsl:copy-of select="@id"/>
          
          <!-- Workaround for FOP bug 48765 - 
               java.lang.NullPointerException in
               org.apache.fop.layoutmgr.inline.InlineStackingLayoutManager.
               getChangedKnuthElements -->

          <xsl:if test="$foProcessor eq 'FOP' and $hyphenate eq 'yes'">
            <xsl:text>&#xFEFF;</xsl:text>
          </xsl:if>
        </fo:inline>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- role-index-entry ================================================== -->

  <xsl:attribute-set name="role-index-entry"
                     use-attribute-sets="compact-block-style">
    <xsl:attribute name="hyphenate">false</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:li[u:has(@class, 'role-index-entry')]">
    <fo:block xsl:use-attribute-sets="role-index-entry">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-index-anchor-refs ============================================ -->

  <xsl:template match="html:span[u:has(@class, 'role-index-anchor-refs')]">
    <xsl:param name="pageNumberSepar" select="''" tunnel="yes"/>
    <xsl:param name="pageRangeSepar" select="''" tunnel="yes"/>

    <xsl:choose>
      <xsl:when test="$foProcessor eq 'XEP'">
        <rx:page-index xmlns:rx="http://www.renderx.com/XSL/Extensions">
          <!-- list-separator="{$pageNumberSepar}" is ignored. -->
          <rx:index-item ref-key="{../@id}"
                         range-separator="{$pageRangeSepar}"
                         merge-subsequent-page-numbers="false"
                         link-back="true"/>
        </rx:page-index>
      </xsl:when>

      <xsl:when test="$foProcessor eq 'AHF'">
        <!-- Other value is "merge". Leave separate because of the links. -->
        <fo:index-page-citation-list
          merge-sequential-page-numbers="leave-separate">
          <fo:index-page-citation-list-separator><xsl:value-of 
            select="$pageNumberSepar"/></fo:index-page-citation-list-separator>

          <fo:index-page-citation-range-separator><xsl:value-of
            select="$pageRangeSepar"/></fo:index-page-citation-range-separator>

          <fo:index-key-reference ref-index-key="{../@id}"
                                  page-number-treatment="link"/>
        </fo:index-page-citation-list>
      </xsl:when>

      <xsl:otherwise>
        <!-- Transparent. -->
        <xsl:apply-templates/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- Transparent role-index-range ====================================== -->

  <xsl:template match="html:span[u:has(@class, 'role-index-range')]">
    <xsl:apply-templates/>
  </xsl:template>

  <!-- role-index-anchor-ref ============================================= -->

  <xsl:attribute-set name="role-index-anchor-ref">
  </xsl:attribute-set>

  <xsl:template match="html:a[u:has(@class, 'role-index-anchor-ref')]">
    <xsl:variable name="ref" select="substring-after(@href, '#')"/>
    
    <fo:basic-link internal-destination="{$ref}"
                   xsl:use-attribute-sets="role-index-anchor-ref">
      <xsl:call-template name="cssStyles">
        <xsl:with-param name="select" select="'inline'"/>
      </xsl:call-template>
      <xsl:call-template name="commonAttributes"/>

      <fo:page-number-citation ref-id="{$ref}"/>
    </fo:basic-link>
  </xsl:template>

  <!-- role-part ========================================================= -->

  <xsl:attribute-set name="role-part">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-part')]">
    <xsl:variable name="subsections" 
                  select="html:section[u:has(@class, 'role-chapter')]"/>

    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-part">
          <xsl:call-template name="rootBlockCSSStyles"/>
          <xsl:call-template name="commonAttributes"/>

          <xsl:apply-templates  select="* except $subsections"/>
        </fo:block>

        <xsl:call-template name="endOfBody"/>
      </fo:flow>
    </fo:page-sequence>

    <xsl:apply-templates select="$subsections"/>
  </xsl:template>

  <xsl:template name="endOfBody">
    <xsl:variable name="bodyDivisions" 
      select="//html:section[u:has(@class, 
                                   ('role-part', 'role-chapter',
                                    'role-appendices', 'role-appendix'))]"/>
    <xsl:variable name="pos" select="u:indexOfNode($bodyDivisions, .)"/>

    <xsl:if test="number($pos) eq count($bodyDivisions)">
      <fo:block id="__EOBO" hyphenate="false"/>
    </xsl:if>
  </xsl:template>

  <!-- role-part-title =================================================== -->

  <xsl:attribute-set name="role-part-title" use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-part-title')]">
    <fo:block xsl:use-attribute-sets="role-part-title">
      <xsl:call-template name="processChapterTitle"/>
    </fo:block>
  </xsl:template>

  <xsl:template name="processChapterTitle">
    <xsl:variable name="marker" select="normalize-space(.)"/>
    <xsl:call-template name="xfcOutlineLevel"/>
    <xsl:call-template name="cssStyles"/>
    <xsl:call-template name="commonAttributes"/>

    <fo:marker marker-class-name="chapter-title">
      <xsl:value-of select="$marker"/>
    </fo:marker>
    <fo:marker marker-class-name="section1-title">
      <xsl:value-of select="$marker"/>
    </fo:marker>
    <fo:marker marker-class-name="division-title">
      <xsl:value-of select="$marker"/>
    </fo:marker>

    <xsl:apply-templates/>
  </xsl:template>

  <!-- role-chapter ====================================================== -->

  <xsl:attribute-set name="role-chapter">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-chapter')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-chapter">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfBody"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-chapter-title ================================================ -->

  <xsl:attribute-set name="role-chapter-title"
                     use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-chapter-title')]">
    <fo:block xsl:use-attribute-sets="role-chapter-title">
      <xsl:call-template name="processChapterTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-appendices =================================================== -->

  <xsl:attribute-set name="role-appendices">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-appendices')]">
    <xsl:variable name="subsections" 
                  select="html:section[u:has(@class, 'role-appendix')]"/>

    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-appendices">
          <xsl:call-template name="rootBlockCSSStyles"/>
          <xsl:call-template name="commonAttributes"/>

          <xsl:apply-templates  select="* except $subsections"/>
        </fo:block>

        <xsl:call-template name="endOfBody"/>
      </fo:flow>
    </fo:page-sequence>

    <xsl:apply-templates select="$subsections"/>
  </xsl:template>

  <!-- role-appendices-title ============================================= -->

  <xsl:attribute-set name="role-appendices-title" 
                     use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-appendices-title')]">
    <fo:block xsl:use-attribute-sets="role-appendices-title">
      <xsl:call-template name="processChapterTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-appendix ===================================================== -->

  <xsl:attribute-set name="role-appendix">
  </xsl:attribute-set>

  <xsl:template match="html:section[u:has(@class, 'role-appendix')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-appendix">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfBody"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <!-- role-appendix-title =============================================== -->

  <xsl:attribute-set name="role-appendix-title"
                     use-attribute-sets="h1 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h1[u:has(@class, 'role-appendix-title')]">
    <fo:block xsl:use-attribute-sets="role-appendix-title">
      <xsl:call-template name="processChapterTitle"/>
    </fo:block>
  </xsl:template>

  <!-- role-section1 ===================================================== -->

  <xsl:attribute-set name="role-section1">
  </xsl:attribute-set>

  <xsl:template match="html:div[@class=('role-frontmatter','role-backmatter')]/
                       html:section[u:has(@class, 'role-section1')]">
    <fo:page-sequence>
      <xsl:call-template name="configurePageSequence"/>

      <fo:flow flow-name="xsl-region-body">
        <fo:block xsl:use-attribute-sets="base-style role-section1">
          <xsl:call-template name="processRootBlock"/>
        </fo:block>

        <xsl:call-template name="endOfFrontBackMatter"/>
      </fo:flow>
    </fo:page-sequence>
  </xsl:template>

  <xsl:template match="html:section[u:has(@class,
                                          ('role-chapter','role-appendix'))]/
                       html:section[u:has(@class, 'role-section1')]">
    <fo:block xsl:use-attribute-sets="role-section1">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-section2,3,etc processed normally ============================ -->

  <!-- role-section1-title =============================================== -->

  <xsl:attribute-set name="role-section1-title"
                     use-attribute-sets="h2 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h2[u:has(@class, 'role-section1-title')]">
    <fo:block xsl:use-attribute-sets="role-section1-title">
      <xsl:call-template name="xfcOutlineLevel"/>
      <xsl:call-template name="cssStyles"/>
      <xsl:call-template name="commonAttributes"/>

      <xsl:variable name="marker" select="normalize-space(.)"/>
      <fo:marker marker-class-name="section1-title">
        <xsl:value-of select="$marker"/>
      </fo:marker>
      <fo:marker marker-class-name="division-title">
        <xsl:value-of select="$marker"/>
      </fo:marker>

      <xsl:apply-templates/>
    </fo:block>
  </xsl:template>

  <!-- role-section2-title =============================================== -->

  <xsl:attribute-set name="role-section2-title"
                     use-attribute-sets="h3 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h3[u:has(@class, 'role-section2-title')]">
    <fo:block xsl:use-attribute-sets="role-section2-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="2"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <xsl:template name="processSectionTitle">
    <xsl:param name="level" select="1"/>

    <xsl:call-template name="xfcOutlineLevel">
      <xsl:with-param name="level" select="$level"/>
    </xsl:call-template>
    <xsl:call-template name="cssStyles"/>
    <xsl:call-template name="commonAttributes"/>

    <fo:marker marker-class-name="division-title">
      <xsl:value-of select="normalize-space(.)"/>
    </fo:marker>

    <xsl:apply-templates/>
  </xsl:template>

  <!-- role-section3-title =============================================== -->

  <xsl:attribute-set name="role-section3-title"
                     use-attribute-sets="h4 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h4[u:has(@class, 'role-section3-title')]">
    <fo:block xsl:use-attribute-sets="role-section3-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="3"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section4-title =============================================== -->

  <xsl:attribute-set name="role-section4-title"
                     use-attribute-sets="h5 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h5[u:has(@class, 'role-section4-title')]">
    <fo:block xsl:use-attribute-sets="role-section4-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="4"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section5-title =============================================== -->

  <xsl:attribute-set name="role-section5-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section5-title')]">
    <fo:block xsl:use-attribute-sets="role-section5-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="5"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section6-title =============================================== -->

  <xsl:attribute-set name="role-section6-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section6-title')]">
    <fo:block xsl:use-attribute-sets="role-section6-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="6"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section7-title =============================================== -->

  <xsl:attribute-set name="role-section7-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section7-title')]">
    <fo:block xsl:use-attribute-sets="role-section7-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="7"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section8-title =============================================== -->

  <xsl:attribute-set name="role-section8-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section8-title')]">
    <fo:block xsl:use-attribute-sets="role-section8-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="8"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section9-title =============================================== -->

  <xsl:attribute-set name="role-section9-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section9-title')]">
    <fo:block xsl:use-attribute-sets="role-section9-title">
      <xsl:call-template name="processSectionTitle">
        <xsl:with-param name="level" select="9"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-section-title ================================================ -->

  <xsl:attribute-set name="role-section-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-section-title')]">
    <fo:block xsl:use-attribute-sets="role-section-title">
      <xsl:call-template name="cssStyles"/>
      <xsl:call-template name="commonAttributes"/>

      <fo:marker marker-class-name="division-title">
        <xsl:value-of select="normalize-space(.)"/>
      </fo:marker>

      <xsl:apply-templates/>
    </fo:block>
  </xsl:template>

  <!-- div.role-content processed normally =============================== -->

  <!-- role-local-toc processed normally  ================================ -->

  <!-- role-local-toc-title ============================================== -->

  <xsl:attribute-set name="role-local-toc-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-local-toc-title')]">
    <fo:block xsl:use-attribute-sets="role-local-toc-title">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-related processed normally  ================================== -->

  <!-- role-related-title ================================================ -->

  <xsl:attribute-set name="role-related-title"
                     use-attribute-sets="h6 title-style">
  </xsl:attribute-set>

  <xsl:template match="html:h6[u:has(@class, 'role-related-title')]">
    <fo:block xsl:use-attribute-sets="role-related-title">
      <xsl:call-template name="processBlock"/>
    </fo:block>
  </xsl:template>

  <!-- role-note ========================================================= -->

  <xsl:attribute-set name="role-note" use-attribute-sets="blockquote">
    <xsl:attribute name="keep-together.within-column">always</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-note')]">
    <fo:block xsl:use-attribute-sets="role-note">
      <xsl:call-template name="processNote"/>
    </fo:block>
  </xsl:template>

  <xsl:template name="processNote">
    <xsl:param name="type" select="'note'"/>

    <xsl:if test="$foProcessor eq 'XFC'">
      <xsl:attribute name="xfc:render-as-table" 
        xmlns:xfc="http://www.xmlmind.com/foconverter/xsl/extensions"
        select="if ($xfcRenderAsTable = 'admonition') then 'true' else 'false'"/>
    </xsl:if>

    <xsl:call-template name="cssStyles"/>
    <xsl:call-template name="commonAttributes"/>

    <xsl:choose>
      <xsl:when test="$use-note-icon eq 'yes'">
        <xsl:choose>
          <xsl:when test="$foProcessor eq 'XFC'">
            <xsl:call-template name="processNoteWithIcon2">
              <xsl:with-param name="type" select="$type"/>
            </xsl:call-template>
          </xsl:when>
          <xsl:otherwise>
            <xsl:call-template name="processNoteWithIcon">
              <xsl:with-param name="type" select="$type"/>
            </xsl:call-template>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>

      <xsl:otherwise>
        <xsl:call-template name="processNoteContent">
          <xsl:with-param name="type" select="$type"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:attribute-set name="note-with-icon">
    <xsl:attribute name="provisional-label-separation"
                   select="concat($note-icon-width, ' * 0.5')"/>
    <xsl:attribute name="provisional-distance-between-starts"
                   select="concat($note-icon-width, ' * 1.5')"/>
  </xsl:attribute-set>

  <xsl:template name="processNoteWithIcon">
    <xsl:param name="type" select="'note'"/>

    <fo:list-block xsl:use-attribute-sets="note-with-icon">
      <fo:list-item relative-align="before">
        <fo:list-item-label end-indent="label-end()">
          <fo:block text-align="end">
            <xsl:call-template name="processNoteIcon">
              <xsl:with-param name="type" select="$type"/>
            </xsl:call-template>
          </fo:block>
        </fo:list-item-label>

        <fo:list-item-body start-indent="body-start()">
          <fo:block>
            <xsl:call-template name="processNoteContent">
              <xsl:with-param name="type" select="$type"/>
            </xsl:call-template>
          </fo:block>
        </fo:list-item-body>
      </fo:list-item>
    </fo:list-block>
  </xsl:template>

  <xsl:template name="processNoteIcon">
    <xsl:param name="type" select="'note'"/>

    <fo:external-graphic src="{concat('url(',$xslResourcesDir,$type,'.svg)')}"
                         content-width="{$note-icon-width}"
                         content-height="{$note-icon-height}">
      <xsl:attribute name="role">
        <xsl:call-template name="localize">
          <xsl:with-param name="message" select="$type"/>
        </xsl:call-template>
      </xsl:attribute>
    </fo:external-graphic>
  </xsl:template>

  <xsl:attribute-set name="note-layout">
  </xsl:attribute-set>

  <xsl:attribute-set name="note-layout-icon">
    <xsl:attribute name="padding-right"
                   select="concat($note-icon-width, ' * 0.5')"/>
  </xsl:attribute-set>

  <xsl:attribute-set name="note-layout-content">
  </xsl:attribute-set>

  <xsl:template name="processNoteWithIcon2">
    <xsl:param name="type" select="'note'"/>

    <fo:table table-layout="auto" xsl:use-attribute-sets="note-layout">
      <fo:table-body>
        <fo:table-row>
          <fo:table-cell start-indent="0"
                         xsl:use-attribute-sets="note-layout-icon">
            <fo:block>
              <xsl:call-template name="processNoteIcon">
                <xsl:with-param name="type" select="$type"/>
              </xsl:call-template>
            </fo:block>
          </fo:table-cell>

          <fo:table-cell start-indent="0"
                         xsl:use-attribute-sets="note-layout-content">
            <fo:block>
              <xsl:call-template name="processNoteContent">
                <xsl:with-param name="type" select="$type"/>
              </xsl:call-template>
            </fo:block>
          </fo:table-cell>
        </fo:table-row>
      </fo:table-body>
    </fo:table>
  </xsl:template>

  <xsl:attribute-set name="note-label" use-attribute-sets="h6">
    <xsl:attribute name="space-before">0em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template name="processNoteContent">
    <xsl:param name="type" select="'note'"/>

    <xsl:if test="$use-note-label eq 'yes' and
                  empty(./*[position() eq 1 and
                            (self::html:h1 or
                             self::html:h2 or
                             self::html:h3 or
                             self::html:h4 or
                             self::html:h5 or
                             self::html:h6)])">
      <fo:block xsl:use-attribute-sets="note-label">
        <xsl:call-template name="localize">
          <xsl:with-param name="message" select="$type"/>
        </xsl:call-template>
      </fo:block>
    </xsl:if>

    <xsl:apply-templates/>
  </xsl:template>

  <!-- role-attention ==================================================== -->

  <xsl:attribute-set name="role-attention" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-attention')]">
    <fo:block xsl:use-attribute-sets="role-attention">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'attention'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-caution ====================================================== -->

  <xsl:attribute-set name="role-caution" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-caution')]">
    <fo:block xsl:use-attribute-sets="role-caution">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'caution'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-danger ======================================================= -->

  <xsl:attribute-set name="role-danger" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-danger')]">
    <fo:block xsl:use-attribute-sets="role-danger">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'danger'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-fastpath ===================================================== -->

  <xsl:attribute-set name="role-fastpath" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-fastpath')]">
    <fo:block xsl:use-attribute-sets="role-fastpath">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'fastpath'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-important ==================================================== -->

  <xsl:attribute-set name="role-important" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-important')]">
    <fo:block xsl:use-attribute-sets="role-important">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'important'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-notice ======================================================= -->

  <xsl:attribute-set name="role-notice" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-notice')]">
    <fo:block xsl:use-attribute-sets="role-notice">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'notice'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-remember ===================================================== -->

  <xsl:attribute-set name="role-remember" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-remember')]">
    <fo:block xsl:use-attribute-sets="role-remember">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'remember'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-restriction ================================================== -->

  <xsl:attribute-set name="role-restriction" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-restriction')]">
    <fo:block xsl:use-attribute-sets="role-restriction">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'restriction'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-tip ========================================================== -->

  <xsl:attribute-set name="role-tip" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-tip')]">
    <fo:block xsl:use-attribute-sets="role-tip">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'tip'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-trouble ====================================================== -->

  <xsl:attribute-set name="role-trouble" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-trouble')]">
    <fo:block xsl:use-attribute-sets="role-trouble">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'trouble'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- role-warning ====================================================== -->

  <xsl:attribute-set name="role-warning" use-attribute-sets="role-note">
  </xsl:attribute-set>

  <xsl:template match="html:blockquote[u:has(@class, 'role-warning')]">
    <fo:block xsl:use-attribute-sets="role-warning">
      <xsl:call-template name="processNote">
        <xsl:with-param name="type" select="'warning'"/>
      </xsl:call-template>
    </fo:block>
  </xsl:template>

  <!-- div.role-footnotes never traversed (see below) ==================== -->
  <!-- Hence hr.role-footnote-separator not processed. -->

  <xsl:template match="html:div[u:has(@class, 'role-footnotes')]"/>

  <!-- role-footnote-ref, role-footnote ================================== -->

  <xsl:key name="footnoteRef"
           match="html:a[u:has(@class, 'role-footnote-ref') and 
                         starts-with(@href, '#')]" 
           use="substring-after(@href, '#')"/>

  <xsl:template match="html:a[u:has(@class, 'role-footnote-ref')]">
    <xsl:variable name="id" select="substring-after(@href, '#')"/>

    <xsl:variable name="footnote"
                  select="//html:div[u:has(@class, 'role-footnote') and 
                                     @id eq $id]"/>

    <xsl:choose>
      <xsl:when test="exists($footnote)">
        <xsl:choose>
          <xsl:when test="key('footnoteRef', $id)[1] is .">
            <!-- First reference to the footnote. -->
            <fo:footnote>
              <xsl:call-template name="footnoteCallout">
                <xsl:with-param name="footnote" select="$footnote[1]"/>
              </xsl:call-template>

              <xsl:for-each select="$footnote[1]">
                <xsl:call-template name="footnoteBody"/>
              </xsl:for-each>
            </fo:footnote>
          </xsl:when>

          <xsl:otherwise>
            <!-- Other reference to the footnote.
                 (Possibly found in another page.) -->
            <xsl:call-template name="footnoteCallout">
              <xsl:with-param name="footnote" select="$footnote[1]"/>
            </xsl:call-template>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>
        <fo:inline xsl:use-attribute-sets="role-footnote-number">???</fo:inline>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:attribute-set name="role-footnote-number">
    <xsl:attribute name="baseline-shift">super</xsl:attribute>
    <xsl:attribute name="font-size">smaller</xsl:attribute>
    <xsl:attribute name="font-weight">normal</xsl:attribute>
    <xsl:attribute name="font-style">normal</xsl:attribute>
    <xsl:attribute name="font-variant">normal</xsl:attribute>
    <xsl:attribute name="color">black</xsl:attribute>
    <xsl:attribute name="text-decoration">none</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template name="footnoteCallout">
    <xsl:param name="footnote" select="()"/>

    <fo:inline xsl:use-attribute-sets="role-footnote-number">
      <xsl:value-of
        select="$footnote//html:a[u:has(@class, 'role-footnote-number')]"/>
    </fo:inline>
  </xsl:template>

  <xsl:attribute-set name="reset-all-styles" use-attribute-sets="base-style">
    <xsl:attribute name="start-indent">0</xsl:attribute>
    <xsl:attribute name="linefeed-treatment">treat-as-space</xsl:attribute>
    <xsl:attribute name="white-space-collapse">true</xsl:attribute>
    <xsl:attribute
     name="white-space-treatment">ignore-if-surrounding-linefeed</xsl:attribute>
    <xsl:attribute name="wrap-option">wrap</xsl:attribute>
    <xsl:attribute name="text-align">left</xsl:attribute>
    <xsl:attribute name="font-weight">normal</xsl:attribute>
    <xsl:attribute name="font-style">normal</xsl:attribute>
    <xsl:attribute name="font-variant">normal</xsl:attribute>
    <xsl:attribute name="color">black</xsl:attribute>
  </xsl:attribute-set>

  <xsl:attribute-set name="role-footnote"
                     use-attribute-sets="compact-block-style">
    <xsl:attribute name="font-size">0.917em</xsl:attribute>
    <xsl:attribute name="line-height" 
                   select="concat($base-line-height, ' div 0.917')"/>
    <xsl:attribute
        name="provisional-distance-between-starts">2em</xsl:attribute>
    <xsl:attribute name="provisional-label-separation">0.1em</xsl:attribute>
  </xsl:attribute-set>

  <xsl:template name="footnoteBody">
    <fo:footnote-body xsl:use-attribute-sets="reset-all-styles">
      <fo:list-block xsl:use-attribute-sets="role-footnote">
        <xsl:call-template name="cssStyles"/>

        <!-- FOP does not honor relative-align="baseline". XEP does. -->
        <fo:list-item relative-align="baseline">
          <fo:list-item-label end-indent="label-end()">
            <fo:block text-align="end">
              <xsl:text>&#xA0;</xsl:text>
              <xsl:call-template name="footnoteCallout">
                <xsl:with-param name="footnote" select="."/>
              </xsl:call-template>
            </fo:block>
          </fo:list-item-label>

          <fo:list-item-body start-indent="body-start()">
            <fo:block>
              <xsl:call-template name="commonAttributes"/>
              <xsl:apply-templates/>
            </fo:block>
          </fo:list-item-body>
        </fo:list-item>
      </fo:list-block>
    </fo:footnote-body>
  </xsl:template>

  <xsl:template match="html:a[u:has(@class, 'role-footnote-number')]"/>

</xsl:stylesheet>
